grammar Basic;

program: expr;

expr
    : PAROPEN expr PARCLOSE #parens
    | BRAOPEN expr BRACLOSE #block
    | IN PAROPEN variable (LISTDELIM variable)* PARCLOSE #read
    | OUT PAROPEN expr (LISTDELIM expr)* PARCLOSE #write
    | IFOPEN cond=expr IFTRUE onTrue=expr (IFFALSE onFalse=expr)? IFCLOSE #if
    | WHILEOPEN cond=expr WHILETRUE onTrue=expr WHILECLOSE #while
    | op=(PLUS|MINUS|NOT) expr #prefix1
    | lhs=expr op=(MULTIPLY|DIVIDE|MODULO) rhs=expr #infix1
    | lhs=expr op=(PLUS|MINUS) rhs=expr #infix2
    | lhs=expr op=(LT|LEQ|GTE|GT|EQ|NEQ) rhs=expr #infix3
    | lhs=expr AND rhs=expr #infix4
    | lhs=expr OR rhs=expr #infix5
    | DECLARE type IDENTIFIER #declare
    | <assoc=right> variable ASSIGN expr #assign
    | expr (COMPOUND expr?)+ #compound
    | variable #var
    | LITERAL10 #number
    | CHAR #char
    | (TRUE|FALSE) #bool
    ;

type: staticType=(INTTYPE | BOOLTYPE | CHARTYPE) | variable;
variable: IDENTIFIER;

PAROPEN: '(';
PARCLOSE: ')';
BRAOPEN: '{';
BRACLOSE: '}';
IN: 'read';
OUT: 'print';
IFOPEN: 'if';
IFTRUE: 'then';
IFFALSE: 'else';
IFCLOSE: 'fi';
WHILEOPEN: 'while';
WHILETRUE: 'do';
WHILECLOSE: 'od';

PLUS: '+';
MINUS: '-';
NOT: '!';
MULTIPLY: '*';
DIVIDE: '/';
MODULO: '%';
LEQ: '<=';
GTE: '>=';
NEQ: '<>';
EQ: '==';
LT: '<';
GT: '>';
AND: '&&';
OR: '||';

DECLARE: 'var';
INTTYPE: 'int';
BOOLTYPE: 'bool';
CHARTYPE: 'char';
ASSIGN: ':=';
COMPOUND: ';';
LISTDELIM: ',';

CHAR: '\'' . '\'';
TRUE: 'true';
FALSE: 'false';
IDENTIFIER: [A-Za-z_][A-Za-z0-9_]*;
LITERAL10: ('0' | [1-9] [0-9]*);

WHITESPACE: [\t\r\n ] -> skip;
LINECOMMENT: '//' ~[\r\n]* -> skip;
BLOCKCOMMENT: '/*' .*? '*/' -> skip;
